;;; guile-assimp, foreign interface to libassimp
;;; Copyright (C) 2014 by Javier Sancho Fernandez <jsf at jsancho dot org>
;;;
;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (assimp assimp)
  #:use-module (ice-9 iconv)
  #:use-module (rnrs bytevectors)
  #:use-module (system foreign))

(define libassimp (dynamic-link "libassimp"))

(define aiImportFile
  (pointer->procedure '*
		      (dynamic-func "aiImportFile" libassimp)
		      (list '* unsigned-int)))


;;; Type Generation

(define-syntax define-type
  (lambda (x)
    (define (mk-string . args)
      (string-concatenate
       (map (lambda (a)
	      (if (string? a)
		  a
		  (symbol->string (syntax->datum a))))
	    args)))
    (define (mk-symbol . args)
      (datum->syntax x
        (string->symbol
	 (apply mk-string args))))
    (syntax-case x ()
      ((_ name (field field-proc) ...)
       (with-syntax ((type? (mk-symbol #'name "?"))
		     (wrap-type (mk-symbol "wrap-" #'name))
		     (unwrap-type (mk-symbol "unwrap-" #'name))
		     (output-string (mk-string "#<" #'name " ~x>"))
		     (type-contents (mk-symbol #'name "-contents")))
         #'(begin
	     (define-wrapped-pointer-type name
	       type?
	       wrap-type unwrap-type
	       (lambda (x p)
		 (format p output-string
			 (pointer-address (unwrap-type x)))))
	     (define (type-contents wrapped)
	       (let ((unwrapped (unwrap-type wrapped)))
		 (cond ((= (pointer-address unwrapped) 0)
			'())
		       (else
			(filter
			 (lambda (f)
			   (not (null? (cdr f))))
			 (list (cons 'field (field-proc unwrapped))
			       ...))))))))))))


(define (bv-uint-ref pointer index)
  (bytevector-uint-ref
   (pointer->bytevector pointer 4 index)
   0
   (native-endianness)
   4))

(define (get-aiString index)
  (lambda (pointer)
    (let* ((length (bv-uint-ref pointer index))
	   (data (pointer->bytevector pointer length (+ index 4))))
      (bytevector->string data (fluid-ref %default-port-encoding)))))

(define* (get-pointer index #:optional wrap-proc)
  (lambda (pointer)
    (let ((p (bv-uint-ref pointer index)))
      (cond ((= p 0) '())
	    (else
	     (let ((p2 (make-pointer p)))
	       (list
		(cond (wrap-proc (wrap-proc p2))
		      (else p2)))))))))

(define* (get-pointer-of-pointers-procedure num-index root-index #:optional wrap-proc)
  (lambda (pointer)
    (let* ((num (bv-uint-ref pointer num-index))
	   (rootp (make-pointer (bv-uint-ref pointer root-index))))
      (let loop ((i 0))
	(cond ((= i num)
	       '())
	      (else
	       (let* ((p (make-pointer (bv-uint-ref rootp (* 4 i))))
		      (wp (if wrap-proc (wrap-proc p) p)))
		 (cons wp (loop (+ i 1))))))))))

;;; Scenes

(define-type scene
  (flags (lambda (p) (bv-uint-ref p 0)))
  (root-node (lambda (p) (wrap-node (make-pointer (bv-uint-ref p 4)))))
  (meshes (get-pointer-of-pointers-procedure 8 12 wrap-mesh))
  (materials (get-pointer-of-pointers-procedure 16 20))
  (animations (get-pointer-of-pointers-procedure 24 28))
  (textures (get-pointer-of-pointers-procedure 32 36))
  (lights (get-pointer-of-pointers-procedure 40 44))
  (cameras (get-pointer-of-pointers-procedure 48 52)))

(define (load-scene filename flags)
  (wrap-scene
   (aiImportFile (string->pointer filename)
		 flags)))

(export load-scene
	unwrap-scene
	scene?
	scene-contents)


;;; Nodes

(define-type node
  (name (get-aiString 0))
  (transformation (lambda (p) (array->list (pointer->bytevector p 16 1028 'f32))))
  (parent (get-pointer 1092 wrap-node))
  (children (get-pointer-of-pointers-procedure 1096 1100 wrap-node))
  (meshes (get-pointer-of-pointers-procedure 1104 1108 wrap-mesh)))

(export node?
	node-contents)


;;; Meshes

(define-type mesh
  (num-primitive-types (lambda (p) (bv-uint-ref p 0)))
  (vertices (get-pointer-of-pointers-procedure 4 12))
  (faces (get-pointer-of-pointers-procedure 8 124))
  (normals (lambda (p) (bv-uint-ref p 16)))
  (tangents (lambda (p) (bv-uint-ref p 20)))
  (bitangents (lambda (p) (bv-uint-ref p 24)))
  (colors (lambda (p) (bv-uint-ref p 28))) ;AI_MAX_NUMBER_OF_COLOR_SETS
  (texture-coords (lambda (p) (bv-uint-ref p 60))) ;AI_MAX_NUMBER_OF_TEXTURECOORDS
  (num-uv-components (lambda (p) (bv-uint-ref p 92))) ;AI_MAX_NUMBER_OF_TEXTURECOORDS
  (bones (get-pointer-of-pointers-procedure 128 132))
  (material-index (lambda (p) (bv-uint-ref p 136))))

(export mesh?
	mesh-contents)
